;; Generated package description from logview.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "logview" "0.19.4snapshot0.20260218.201306" "Major mode for viewing log files" '((emacs "25.1") (datetime "0.8") (extmap "1.0") (compat "29")) :commit "50cac3f726abd4e7872b8fa36ede72550ce66a64" :authors '(("Paul Pogonyshev" . "pogonyshev@gmail.com")) :maintainer '("Paul Pogonyshev" . "pogonyshev@gmail.com") :keywords '("files" "tools") :url "https://github.com/doublep/logview")
